# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.field import Field, FieldException, Function, JsonRelated
from anyblok.column import Integer, String, Json
from anyblok import Declarations
from sqlalchemy import func, types


Model = Declarations.Model
register = Declarations.register


class OneField(Field):
    pass


class TestField(TestCase):

    def test_forbid_instance(self):
        try:
            Field()
            self.fail("Field mustn't be instanciated")
        except FieldException:
            pass

    def test_without_label(self):
        field = OneField()
        field.get_sqlalchemy_mapping(None, None, 'a_field', None)
        self.assertEqual(field.label, 'A field')


def field_without_name():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        field = OneField()


class TestField2(DBTestCase):

    def test_field_without_name(self):
        self.init_registry(field_without_name)

    def define_field_function(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            first_name = String()
            last_name = String()
            name = Function(fget='fget', fset='fset', fdel='fdel',
                            fexpr='fexpr')

            def fget(self):
                return '{0} {1}'.format(self.first_name, self.last_name)

            def fset(self, value):
                self.first_name, self.last_name = value.split(' ', 1)

            def fdel(self):
                self.first_name = self.last_name = None

            @classmethod
            def fexpr(cls):
                return func.concat(cls.first_name, ' ', cls.last_name)

    def test_field_function_fget(self):
        registry = self.init_registry(self.define_field_function)
        t = registry.Test.insert(first_name='Jean-Sebastien',
                                 last_name='SUZANNE')
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')
        t = registry.Test.query().first()
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')

    def test_field_function_fset(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                _name = String()
                name = Function(fget='fget', fset='fset', fdel='fdel',
                                fexpr='fexpr')

                def fget(self):
                    return self._name

                def fset(self, value):
                    self._name = value

                def fdel(self):
                    self._name = None

                @classmethod
                def fexpr(cls):
                    return cls._name

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert(name='Jean-Sebastien SUZANNE')
        self.assertEqual(t._name, 'Jean-Sebastien SUZANNE')
        t = registry.Test.query().first()
        t.name = 'Mister ANYBLOK'
        self.assertEqual(t._name, 'Mister ANYBLOK')
        t.update(name='Jean-Sebastien SUZANNE')
        self.assertEqual(t._name, 'Jean-Sebastien SUZANNE')

    def test_field_function_fdel(self):
        registry = self.init_registry(self.define_field_function)
        t = registry.Test.insert(first_name='jean-sebastien',
                                 last_name='suzanne')
        del t.name
        self.assertEqual(t.first_name, None)
        self.assertEqual(t.last_name, None)

    def test_field_function_fexpr(self):
        registry = self.init_registry(self.define_field_function)
        registry.Test.insert(first_name='Jean-Sebastien',
                             last_name='SUZANNE')
        t = registry.Test.query().filter(
            registry.Test.name == 'Jean-Sebastien SUZANNE').first()
        self.assertEqual(t.name, 'Jean-Sebastien SUZANNE')

    def test_class_hasattr(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val1 = Integer()
                val2 = Function(fget='fget', fset='fset')

                def fget(self):
                    return 2 * self.val1

        registry = self.init_registry(add_in_registry)
        registry.Test.val2

    def define_field_json_related(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            properties = Json()
            name = JsonRelated(json_column='properties', keys=['name'])

    def define_field_json_related2(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            properties = Json()
            name = JsonRelated(json_column='properties', keys=['sub', 'name'])

    def define_field_json_related_multi_keys(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            properties = Json()
            name = JsonRelated(json_column='properties', keys=['sub', 'name'])

    def test_field_json_related_hasattr(self):
        registry = self.init_registry(self.define_field_json_related)
        registry.Test.name

    def test_field_json_related_get_1(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert()
        self.assertIsNone(t.name)

    def test_field_json_related_get_2(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert(properties={})
        self.assertIsNone(t.name)

    def test_field_json_related_get_3(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert(properties={'name': 'jssuzanne'})
        self.assertEqual(t.name, 'jssuzanne')

    def test_field_json_related_get_4(self):
        registry = self.init_registry(self.define_field_json_related2)
        t = registry.Test.insert(
            properties={'sub': {'name': 'jssuzanne'}})
        self.assertEqual(t.name, 'jssuzanne')

    def test_field_json_related_set_1(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert()
        t.name = 'jssuzanne'
        self.assertEqual(t.properties, {'name': 'jssuzanne'})

    def test_field_json_related_set_2(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert(properties={})
        t.name = 'jssuzanne'
        self.assertEqual(t.properties, {'name': 'jssuzanne'})

    def test_field_json_related_set_3(self):
        registry = self.init_registry(self.define_field_json_related)
        t = registry.Test.insert(properties={'name': 'other'})
        t.name = 'jssuzanne'
        self.assertEqual(t.properties, {'name': 'jssuzanne'})

    def test_field_json_related_set_4(self):
        registry = self.init_registry(self.define_field_json_related2)
        t = registry.Test.insert(properties={'sub': {'name': 'other'}})
        t.name = 'jssuzanne'
        self.assertEqual(t.properties, {'sub': {'name': 'jssuzanne'}})

    def test_field_json_related_exp_1(self):
        registry = self.init_registry(self.define_field_json_related)
        Test = registry.Test
        Test.insert()
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        self.assertFalse(query.count())

    def test_field_json_related_exp_2(self):
        registry = self.init_registry(self.define_field_json_related)
        Test = registry.Test
        Test.insert(properties={})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        self.assertFalse(query.count())

    def test_field_json_related_exp_3(self):
        registry = self.init_registry(self.define_field_json_related)
        Test = registry.Test
        Test.insert(properties={'name': 'jssuzanne'})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        self.assertTrue(query.count())

    def test_field_json_related_exp_4(self):
        registry = self.init_registry(self.define_field_json_related2)
        Test = registry.Test
        Test.insert(properties={'sub': {'name': 'jssuzanne'}})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        self.assertTrue(query.count())

    def test_field_json_related_with_adapter(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(json_column='properties', keys=['name'],
                                   set_adapter=str, get_adapter=int)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertIsNone(t.name)
        t.name = 1
        self.assertEqual(t.properties, {'name': '1'})
        t.properties['name'] = '2'
        self.assertEqual(t.name, 2)

    def test_field_json_related_with_adapter_with_method(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(json_column='properties', keys=['name'],
                                   set_adapter='fromint', get_adapter='toint')

                def fromint(self, value):
                    if value:
                        return str(value)

                    return value

                def toint(self, value):
                    if value:
                        return int(value)

                    return value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertIsNone(t.name)
        t.name = 1
        self.assertEqual(t.properties, {'name': '1'})
        t.properties['name'] = '2'
        self.assertEqual(t.name, 2)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestSystemSequence(BlokTestCase):

    def test_nextval_without_prefix_without_suffix(self):
        Sequence = self.registry.System.Sequence
        seq = Sequence.insert(code='test.sequence')
        number = seq.number
        self.assertEqual(seq.nextval(), str(number + 1))
        self.assertEqual(seq.nextval(), str(number + 2))
        self.assertEqual(seq.nextval(), str(number + 3))

    def test_nextval_without_prefix_without_suffix_two_time(self):
        Sequence = self.registry.System.Sequence
        Sequence.insert(code='test.sequence')
        seq = Sequence.insert(code='test.sequence')
        number = seq.number
        self.assertEqual(seq.nextval(), str(number + 1))
        self.assertEqual(seq.nextval(), str(number + 2))
        self.assertEqual(seq.nextval(), str(number + 3))

    def test_nextval_without_prefix_with_suffix(self):
        Sequence = self.registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', formater="{seq}_suffix")
        number = seq.number
        self.assertEqual(seq.nextval(), '%d_suffix' % (number + 1))
        self.assertEqual(seq.nextval(), '%d_suffix' % (number + 2))
        self.assertEqual(seq.nextval(), '%d_suffix' % (number + 3))

    def test_nextval_with_prefix_without_suffix(self):
        Sequence = self.registry.System.Sequence
        seq = Sequence.insert(code='test.sequence', formater='prefix_{seq}')
        number = seq.number
        self.assertEqual(seq.nextval(), 'prefix_%d' % (number + 1))
        self.assertEqual(seq.nextval(), 'prefix_%d' % (number + 2))
        self.assertEqual(seq.nextval(), 'prefix_%d' % (number + 3))

    def test_nextval_with_prefix_with_suffix(self):
        Sequence = self.registry.System.Sequence
        seq = Sequence.insert(code='test.sequence',
                              formater='prefix_{seq}_suffix')
        number = seq.number
        self.assertEqual(seq.nextval(), 'prefix_%d_suffix' % (number + 1))
        self.assertEqual(seq.nextval(), 'prefix_%d_suffix' % (number + 2))
        self.assertEqual(seq.nextval(), 'prefix_%d_suffix' % (number + 3))

    def test_nextval_by_attribute(self):
        Sequence = self.registry.System.Sequence
        seq = Sequence.insert(code='test.sequence')
        number = seq.number
        self.assertEqual(Sequence.nextvalBy(code=seq.code), str(number + 1))
        self.assertEqual(Sequence.nextvalBy(code=seq.code), str(number + 2))
        self.assertEqual(Sequence.nextvalBy(code=seq.code), str(number + 3))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import ExporterException


class TestExporter(BlokTestCase):

    def test_get_external_id(self):
        Exporter = self.registry.IO.Exporter
        val1 = Exporter.get_external_id(Exporter.__registry_name__)
        val2 = Exporter.get_external_id(Exporter.__registry_name__)
        val1 = val1.split('_')
        val1 = '_'.join([val1[0], str(int(val1[1]) + 1)])
        self.assertEqual(val1, val2)

    def test_get_key_for_mapping(self):
        Exporter = self.registry.IO.Exporter
        Blok = self.registry.System.Blok
        entry = Blok.from_primary_keys(name='anyblok-core')
        val1 = Exporter.get_external_id(Blok.__registry_name__)
        val1 = val1.split('_')
        val1 = '_'.join([val1[0], str(int(val1[1]) + 1)])
        self.assertEqual(Exporter.get_key_mapping(entry), val1)

    def test_get_key_for_mapping_2nd_time(self):
        Exporter = self.registry.IO.Exporter
        Blok = self.registry.System.Blok
        entry = Blok.from_primary_keys(name='anyblok-core')
        val1 = Exporter.get_external_id(Blok.__registry_name__)
        val1 = val1.split('_')
        val1 = '_'.join([val1[0], str(int(val1[1]) + 1)])
        self.assertEqual(Exporter.get_key_mapping(entry), val1)
        self.assertEqual(Exporter.get_key_mapping(entry), val1)

    def test_export_with_entries_doesnt_come_from_model(self):
        Exporter = self.registry.IO.Exporter
        Blok = self.registry.System.Blok
        exporter = Exporter(model=Blok.__registry_name__)
        with self.assertRaises(ExporterException):
            exporter.run([exporter])



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from logging import getLogger
logger = getLogger(__name__)


def load_init_function_from_entry_points(unittest=False):
    """Call all the entry point ``anyblok_pyramid.init`` to update
    the argument setting

    the callable need to have one parametter, it is a dict::

        def init_function(unittest=False):
            ...

    We add the entry point by the setup file::

        setup(
            ...,
            entry_points={
                'anyblok.init': [
                    init_function=path:init_function,
                    ...
                ],
            },
            ...,
        )


    """
    for i in iter_entry_points('anyblok.init'):
        print('AnyBlok Load init: %r' % i)
        i.load()(unittest=unittest)


def configuration_post_load(unittest=False):
    """Call all the entry point ``anyblok_configuration.post_load`` to
    initialize some service in function of the configuration

    the callable need to have one parametter, it is a dict::

        def post_load_function(unittest=False):
            ...

    We add the entry point by the setup file::

        setup(
            ...,
            entry_points={
                'anyblok_configuration.post_load': [
                    post_load_function=path:post_load_function,
                    ...
                ],
            },
            ...,
        )


    """
    for i in iter_entry_points('anyblok_configuration.post_load'):
        logger.info('AnyBlok configuration post load: %r' % i)
        i.load()(unittest=unittest)


def start(processName, entry_points=None,
          useseparator=False, loadwithoutmigration=False, config=None,
          **kwargs):
    """ Function which initialize the application

    ::

        registry = start('My application',
                         entry_points=['AnyBlok'])

    :param processName: Name of the application
    :param version: Version of the application
    :param prompt: Prompt message for the help
    :param entry_points: entry point where load blok
    :param useseparator: boolean, indicate if configuration option are split
        betwwen two application
    :param loadwithoutmigration: if True, any migration operation will do
    :rtype: registry if the database name is in the configuration
    """
    from .blok import BlokManager
    from .config import Configuration
    from .registry import RegistryManager

    load_init_function_from_entry_points()
    if config is None:
        config = {}

    Configuration.load(processName,
                       useseparator=useseparator, **config)

    configuration_post_load()
    if entry_points:
        BlokManager.load(entry_points=entry_points)
    else:
        BlokManager.load()

    db_name = Configuration.get('db_name')
    logger.debug("start(): db_name=%r", db_name)
    if not db_name:
        logger.warning("start(): no database name in configuration, "
                       "bailing out")
        return None

    registry = RegistryManager.get(
        db_name, loadwithoutmigration=loadwithoutmigration, **kwargs)
    registry.commit()
    return registry


from .declarations import Declarations  # noqa
from . import core  # noqa
from . import model  # noqa
from . import mixin  # noqa
from .authorization import binding  # noqa
from .imp import reload_module_if_blok_is_reloading  # noqa



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import BlokManager
from anyblok.declarations import Declarations, listen, classmethod_cache
from anyblok.column import String, Integer, Selection
from anyblok.field import Function
from logging import getLogger
from os.path import join, isfile


logger = getLogger(__name__)
register = Declarations.register
System = Declarations.Model.System


@register(System)
class Blok:

    STATES = {
        'uninstalled': 'Uninstalled',
        'installed': 'Installed',
        'toinstall': 'To install',
        'toupdate': 'To update',
        'touninstall': 'To uninstall',
        'undefined': 'Undefined',
    }

    name = String(primary_key=True, nullable=False)
    state = Selection(selections=STATES, default='uninstalled', nullable=False)
    author = String()
    order = Integer(default=-1, nullable=False)
    short_description = Function(fget='get_short_description')
    long_description = Function(fget='get_long_description')
    logo = Function(fget='get_logo')
    version = String(nullable=False)
    installed_version = String()

    def get_short_description(self):
        """ fget of the ``short_description`` Column.Selection

        :rtype: the docstring of the blok
        """
        blok = BlokManager.get(self.name)
        if hasattr(blok, '__doc__'):
            return blok.__doc__ or ''

        return ''

    def get_long_description(self):
        """ fget of the ``long_description`` Column.Selection

        :rtype: the readme file of the blok
        """
        blok = BlokManager.get(self.name)
        path = BlokManager.getPath(self.name)
        readme = getattr(blok, 'readme', 'README.rst')
        if readme == '__doc__':
            return blok.__doc__

        file_path = join(path, readme)
        description = ''
        if isfile(file_path):
            with open(file_path, 'r') as fp:
                description = fp.read()

        return description

    def get_logo(self):
        """fget of ``logo`` return the path in the blok of the logo

        :rtype: absolute path or None if unexiste logo
        """
        blok = BlokManager.get(self.name)
        blok_path = BlokManager.getPath(blok.name)
        file_path = join(blok_path, blok.logo)
        if isfile(file_path):
            return file_path

        return None

    def __repr__(self):
        return "%s (%s)" % (self.name, self.state)

    @classmethod
    def list_by_state(cls, *states):
        """ Return the blok name in function of the wanted states

        :param states: list of the state
        :rtype: list if state is a state, dict if the states is a list
        """
        if not states:
            return None

        res = {state: [] for state in states}
        bloks = cls.query().filter(cls.state.in_(states)).order_by(cls.order)
        for blok in bloks.all():
            res[blok.state].append(blok.name)

        if len(states) == 1:
            return res[states[0]]
        return res

    @classmethod
    def update_list(cls):
        """ Populate the bloks list and update the state of existing bloks
        """
        # Do not remove blok because 2 or More AnyBlok api may use the same
        # Database
        for order, blok in enumerate(BlokManager.ordered_bloks):
            b = cls.query().filter(cls.name == blok)
            version = BlokManager.bloks[blok].version
            author = BlokManager.bloks[blok].author
            if b.count():
                b.update(dict(order=order, version=version, author=author))
            else:
                cls.insert(name=blok, order=order, version=version,
                           author=author)

    @classmethod
    def apply_state(cls, *bloks):
        """ Call the rigth method is the blok state change

        .. warning::

            for the uninstallation the method called is ``uninstall_all``

        :param bloks: list of the blok name load by the registry
        """
        for blok in bloks:
            # Make the query in the loop to be sure to keep order
            b = cls.query().filter(cls.name == blok).first()
            if b.state in ('undefined', 'uninstalled', 'toinstall'):
                b.install()
            elif b.state == 'toupdate':
                b.upgrade()

        uninstalled_bloks = cls.query().filter(
            cls.state == 'uninstalled').all().name

        conditional_bloks_to_install = []
        for blok in uninstalled_bloks:
            if cls.check_if_the_conditional_are_installed(blok):
                conditional_bloks_to_install.append(blok)

        if conditional_bloks_to_install:
            for b in conditional_bloks_to_install:
                cls.query().filter(cls.name == b).update(
                    {cls.state: 'toinstall'})

            return True

        return False

    @classmethod
    def uninstall_all(cls, *bloksname):
        """ Search and call the uninstall method for all the uninstalled bloks

        .. warning::

            Use the ``desc order`` to uninstall because we can't uninstall
            a dependancies before

        :param bloksname: list of the blok name to uninstall
        """
        if not bloksname:
            return

        query = cls.query().filter(cls.name.in_(bloksname))
        query = query.order_by(cls.order.desc())
        bloks = query.all()
        if bloks:
            bloks.uninstall()

    @classmethod
    def check_if_the_conditional_are_installed(cls, blok):
        """ Return True if all the conditions to install the blok are satisfied

        :param blok: blok name
        :rtype: boolean
        """
        if blok in BlokManager.bloks:
            conditional = BlokManager.bloks[blok].conditional
            if conditional:
                query = cls.query().filter(cls.name.in_(conditional))
                query = query.filter(
                    cls.state.in_(['installed', 'toinstall', 'toupdate']))
                if len(conditional) == query.count():
                    return True

        return False

    def install(self):
        """ Method to install the blok
        """
        logger.info("Install the blok %r" % self.name)
        self.fire('Update installed blok')
        entry = self.registry.loaded_bloks[self.name]
        entry.update(None)
        self.state = 'installed'
        self.installed_version = self.version

    def upgrade(self):
        """ Method to update the blok
        """
        logger.info("Update the blok %r" % self.name)
        self.fire('Update installed blok')
        entry = self.registry.loaded_bloks[self.name]
        entry.update(self.installed_version)
        self.state = 'installed'
        self.installed_version = self.version

    def uninstall(self):
        """ Method to uninstall the blok
        """
        logger.info("Uninstall the blok %r" % self.name)
        self.fire('Update installed blok')
        entry = BlokManager.bloks[self.name](self.registry)
        entry.uninstall()
        self.state = 'uninstalled'
        self.installed_version = None

    def load(self):
        """ Method to load the blok when the registry is completly loaded
        """
        name = self.name
        blok_cls = BlokManager.get(name)
        if blok_cls is None:
            logger.warning("load(): class of Blok %r not found, "
                           "Blok can't be loaded", name)
            return

        logger.info("Loading Blok %r", name)
        blok_cls(self.registry).load()
        logger.debug("Succesfully loaded Blok %r", name)

    @classmethod
    def load_all(cls):
        """ Load all the installed bloks
        """
        query = cls.query().filter(cls.state == 'installed')
        bloks = query.order_by(cls.order).all()
        if bloks:
            bloks.load()

    @classmethod_cache()
    def is_installed(cls, blok_name):
        return cls.query().filter_by(name=blok_name,
                                     state='installed').count() != 0

    @listen('Model.System.Blok', 'Update installed blok')
    def listen_update_installed_blok(cls):
        cls.registry.System.Cache.invalidate(
            cls.__registry_name__, 'is_installed')
